#
# Revelation 0.2.0 - a password manager for GNOME 2
# http://oss.wired-networks.net/revelation/
#
# Module containing data-related functionality
#
#
# Copyright (c) 2003-2004 Erik Grinaker
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#


import gtk, revelation, gobject, string

ENTRYSTORE_COL_NAME		= 0
ENTRYSTORE_COL_ICON		= 1
ENTRYSTORE_COL_TYPE		= 2
ENTRYSTORE_COL_DESC		= 3
ENTRYSTORE_COL_UPDATED		= 4
ENTRYSTORE_COL_FIELDS		= 5


ENTRY_FOLDER			= "folder"
ENTRY_ACCOUNT_CREDITCARD	= "creditcard"
ENTRY_ACCOUNT_CRYPTOKEY		= "cryptokey"
ENTRY_ACCOUNT_DATABASE		= "database"
ENTRY_ACCOUNT_DOOR		= "door"
ENTRY_ACCOUNT_EMAIL		= "email"
ENTRY_ACCOUNT_FTP		= "ftp"
ENTRY_ACCOUNT_GENERIC		= "generic"
ENTRY_ACCOUNT_PHONE		= "phone"
ENTRY_ACCOUNT_SHELL		= "shell"
ENTRY_ACCOUNT_USENET		= "usenet"
ENTRY_ACCOUNT_WEBSITE		= "website"


FIELD_GENERIC_CERTIFICATE	= "generic-certificate"
FIELD_GENERIC_CODE		= "generic-code"
FIELD_GENERIC_DATABASE		= "generic-database"
FIELD_GENERIC_DOMAIN		= "generic-domain"
FIELD_GENERIC_EMAIL		= "generic-email"
FIELD_GENERIC_HOSTNAME		= "generic-hostname"
FIELD_GENERIC_KEYFILE		= "generic-keyfile"
FIELD_GENERIC_LOCATION		= "generic-location"
FIELD_GENERIC_PASSWORD		= "generic-password"
FIELD_GENERIC_PIN		= "generic-pin"
FIELD_GENERIC_PORT		= "generic-port"
FIELD_GENERIC_URL		= "generic-url"
FIELD_GENERIC_USERNAME		= "generic-username"

FIELD_CREDITCARD_CARDTYPE	= "creditcard-cardtype"
FIELD_CREDITCARD_CARDNUMBER	= "creditcard-cardnumber"
FIELD_CREDITCARD_CCV		= "creditcard-ccv"
FIELD_CREDITCARD_EXPIRYDATE	= "creditcard-expirydate"

FIELD_PHONE_PHONENUMBER		= "phone-phonenumber"

FIELD_TYPE_EMAIL		= "email"
FIELD_TYPE_PASSWORD		= "password"
FIELD_TYPE_TEXT			= "text"
FIELD_TYPE_URL			= "url"

UNDO				= "undo"
REDO				= "redo"


class EntryTypes(object):

	def __init__(self):

		self.entrytype = {

			ENTRY_FOLDER			: {
				"name"		: "Folder",
				"icon"		: revelation.ui.STOCK_FOLDER,
				"fields"	: []
			},

			ENTRY_ACCOUNT_CREDITCARD	: {
				"name"		: "Creditcard",
				"icon"		: revelation.ui.STOCK_ACCOUNT_CREDITCARD,
				"fields"	: [ FIELD_CREDITCARD_CARDTYPE, FIELD_CREDITCARD_CARDNUMBER, FIELD_CREDITCARD_EXPIRYDATE, FIELD_CREDITCARD_CCV, FIELD_GENERIC_PIN ]
			},

			ENTRY_ACCOUNT_CRYPTOKEY		: {
				"name"		: "Crypto Key",
				"icon"		: revelation.ui.STOCK_ACCOUNT_CRYPTOKEY,
				"fields"	: [ FIELD_GENERIC_HOSTNAME, FIELD_GENERIC_CERTIFICATE, FIELD_GENERIC_KEYFILE, FIELD_GENERIC_PASSWORD ]
			},

			ENTRY_ACCOUNT_DATABASE		: {
				"name"		: "Database",
				"icon"		: revelation.ui.STOCK_ACCOUNT_DATABASE,
				"fields"	: [ FIELD_GENERIC_HOSTNAME, FIELD_GENERIC_USERNAME, FIELD_GENERIC_PASSWORD, FIELD_GENERIC_DATABASE ]
			},

			ENTRY_ACCOUNT_DOOR		: {
				"name"		: "Door lock",
				"icon"		: revelation.ui.STOCK_ACCOUNT_DOOR,
				"fields"	: [ FIELD_GENERIC_LOCATION, FIELD_GENERIC_CODE ]
			},

			ENTRY_ACCOUNT_EMAIL		: {
				"name"		: "Email",
				"icon"		: revelation.ui.STOCK_ACCOUNT_EMAIL,
				"fields"	: [ FIELD_GENERIC_EMAIL, FIELD_GENERIC_HOSTNAME, FIELD_GENERIC_USERNAME, FIELD_GENERIC_PASSWORD ]
			},

			ENTRY_ACCOUNT_FTP		: {
				"name"		: "FTP",
				"icon"		: revelation.ui.STOCK_ACCOUNT_FTP,
				"fields"	: [ FIELD_GENERIC_HOSTNAME, FIELD_GENERIC_PORT, FIELD_GENERIC_USERNAME, FIELD_GENERIC_PASSWORD ]
			},

			ENTRY_ACCOUNT_GENERIC		: {
				"name"		: "Generic",
				"icon"		: revelation.ui.STOCK_ACCOUNT,
				"fields"	: [ FIELD_GENERIC_HOSTNAME, FIELD_GENERIC_USERNAME, FIELD_GENERIC_PASSWORD ]
			},

			ENTRY_ACCOUNT_PHONE		: {
				"name"		: "Phone",
				"icon"		: revelation.ui.STOCK_ACCOUNT_PHONE,
				"fields"	: [ FIELD_PHONE_PHONENUMBER, FIELD_GENERIC_PIN ]
			},

			ENTRY_ACCOUNT_SHELL		: {
				"name"		: "Shell",
				"icon"		: revelation.ui.STOCK_ACCOUNT_SHELL,
				"fields"	: [ FIELD_GENERIC_HOSTNAME, FIELD_GENERIC_DOMAIN, FIELD_GENERIC_USERNAME, FIELD_GENERIC_PASSWORD ]
			},

			ENTRY_ACCOUNT_USENET		: {
				"name"		: "Usenet",
				"icon"		: revelation.ui.STOCK_ACCOUNT_USENET,
				"fields"	: [ FIELD_GENERIC_HOSTNAME, FIELD_GENERIC_USERNAME, FIELD_GENERIC_PASSWORD ]
			},

			ENTRY_ACCOUNT_WEBSITE		: {
				"name"		: "Website",
				"icon"		: revelation.ui.STOCK_ACCOUNT_WEBSITE,
				"fields"	: [ FIELD_GENERIC_URL, FIELD_GENERIC_USERNAME, FIELD_GENERIC_PASSWORD ]
			}
		}

		self.fieldtype = {

			FIELD_GENERIC_CODE		: { "name" : "Code",		"type" : FIELD_TYPE_PASSWORD,	"tooltip" : "A code used to provide access to something" },
			FIELD_GENERIC_CERTIFICATE	: { "name" : "Certificate",	"type" : FIELD_TYPE_TEXT,	"tooltip" : "A certificate, such as an X.509 SSL Certificate" },
			FIELD_GENERIC_DATABASE		: { "name" : "Database",	"type" : FIELD_TYPE_TEXT,	"tooltip" : "A database name" },
			FIELD_GENERIC_DOMAIN		: { "name" : "Domain",		"type" : FIELD_TYPE_TEXT,	"tooltip" : "An Internet or logon domain, like amazon.com or a Windows logon domain" },
			FIELD_GENERIC_EMAIL		: { "name" : "Email address",	"type" : FIELD_TYPE_EMAIL,	"tooltip" : "An email address" },
			FIELD_GENERIC_HOSTNAME		: { "name" : "Hostname",	"type" : FIELD_TYPE_TEXT,	"tooltip" : "The name of a computer, like computer.domain.com or MYCOMPUTER" },
			FIELD_GENERIC_KEYFILE		: { "name" : "Key File",	"type" : FIELD_TYPE_TEXT,	"tooltip" : "A key file, used for authentication for example via ssh or to encrypt X.509 certificates" },
			FIELD_GENERIC_LOCATION		: { "name" : "Location",	"type" : FIELD_TYPE_TEXT,	"tooltip" : "A physical location, like office entrance" },
			FIELD_GENERIC_PASSWORD		: { "name" : "Password",	"type" : FIELD_TYPE_PASSWORD,	"tooltip" : "A secret word or character combination used for proving you have access" },
			FIELD_GENERIC_PIN		: { "name" : "PIN",		"type" : FIELD_TYPE_PASSWORD,	"tooltip" : "A Personal Identification Number, a numeric code used for credit cards, phones etc" },
			FIELD_GENERIC_PORT		: { "name" : "Port number",	"type" : FIELD_TYPE_TEXT,	"tooltip" : "A network port number, used to access network services directly" },
			FIELD_GENERIC_URL		: { "name" : "URL",		"type" : FIELD_TYPE_URL,	"tooltip" : "A Uniform Resource Locator, such as a web-site address" },
			FIELD_GENERIC_USERNAME		: { "name" : "Username",	"type" : FIELD_TYPE_TEXT,	"tooltip" : "A name or other identification used to identify yourself" },

			FIELD_CREDITCARD_CARDTYPE	: { "name" : "Card type",	"type" : FIELD_TYPE_TEXT,	"tooltip" : "The type of creditcard, like MasterCard or VISA" },
			FIELD_CREDITCARD_CARDNUMBER 	: { "name" : "Card number",	"type" : FIELD_TYPE_TEXT,	"tooltip" : "The number of a creditcard, usually a 16-digit number" },
			FIELD_CREDITCARD_CCV		: { "name" : "CCV number",	"type" : FIELD_TYPE_TEXT,	"tooltip" : "A Credit Card Verification number, normally a 3-digit code found on the back of a card" },
			FIELD_CREDITCARD_EXPIRYDATE	: { "name" : "Expiry date",	"type" : FIELD_TYPE_TEXT,	"tooltip" : "The month that the credit card validity expires" },

			FIELD_PHONE_PHONENUMBER		: { "name" : "Phone number",	"type" : FIELD_TYPE_TEXT,	"tooltip" : "A telephone number" }
		}


	def field_exists(self, type, field):
		"""Checks if a field exists for a type"""

		return field in self.entrytype[type]["fields"]


	def get_field_data(self, field):
		"""Fetches data for a field type"""

		return self.fieldtype[field]


	def get_list(self):
		"""Fetches a list of entry types"""

		list = self.entrytype.keys()
		list.sort()

		return list


	def get_data(self, type):
		"""Returns data for a type"""

		return self.entrytype[type]


	def type_exists(self, type):
		"""Checks if a type exists"""

		return self.entrytype.has_key(type)



class EntryStore(gtk.TreeStore):
	"""Base class for generic entry store"""

	def __init__(self, entrytypes):
		"""Sets up the entrystore"""

		gtk.TreeStore.__init__(self, gobject.TYPE_STRING, gobject.TYPE_STRING, gobject.TYPE_STRING, gobject.TYPE_STRING, gobject.TYPE_INT, gobject.TYPE_PYOBJECT)
		self.entrytypes = entrytypes


	def add_entry(self, parent, data = None):
		"""Adds an entry to the entry store"""

		iter = self.append(parent)
		self.update_entry(iter, data)
		return iter


	def export_entrystore(self, iters = [], entrystore = None, parent = None):
		"""Exports branches to a new entrystore"""

		if entrystore == None:
			entrystore = EntryStore(EntryTypes())

		for iter in iters:
			child = entrystore.add_entry(parent, self.get_entry(iter))

			if self.iter_has_child(iter):
				for i in range(self.iter_n_children(iter)):
					self.export_entrystore([self.iter_nth_child(iter, i)], entrystore, child)

		return entrystore


	def filter_parents(self, iters):
		"""Removes all decendant iters from a list of iters"""

		parents = []

		for child in iters:
			for parent in iters:
				if self.is_ancestor(parent, child):
					break
			else:
				parents.append(child)

		return parents


	def get_entry(self, iter):
		"""Fetches data for a tree entry"""

		if iter == None:
			return None

		data = {
			"name"		: self.get_value(iter, ENTRYSTORE_COL_NAME),
			"type"		: self.get_value(iter, ENTRYSTORE_COL_TYPE),
			"icon"		: self.get_value(iter, ENTRYSTORE_COL_ICON),
			"description"	: self.get_value(iter, ENTRYSTORE_COL_DESC),
			"updated"	: self.get_value(iter, ENTRYSTORE_COL_UPDATED),
			"fields"	: self.get_value(iter, ENTRYSTORE_COL_FIELDS)
		}

		# fill in any missing keys in the fields dict
		for field in self.entrytypes.get_data(data["type"])["fields"]:
			if not data["fields"].has_key(field):
				data["fields"][field] = ""

		return data


	def get_entry_type(self, iter):
		"""Returns the type of an entry"""

		if iter == None:
			return None
		else:
			return self.get_value(iter, ENTRYSTORE_COL_TYPE)


	def get_iter(self, path):
		"""Returns an iter for a path"""

		if path in [ None, "", () ]:
			return None

		try: iter = gtk.TreeStore.get_iter(self, path)
		except ValueError: iter = None

		return iter


	def get_path(self, iter):
		"""Returns the path to an iter"""

		if iter == None:
			return None
		else:
			return gtk.TreeStore.get_path(self, iter)


	def has_contents(self):
		"""Checks if the entrystore has any contents"""

		return self.iter_n_children(None) > 0


	def import_entrystore(self, entrystore, parent = None, sibling = None, importiter = None):
		"""Imports entries from a different entrystore"""

		baseiters = []

		for i in range(entrystore.iter_n_children(importiter)):
			importchild = entrystore.iter_nth_child(importiter, i)

			if sibling == None:
				childiter = self.add_entry(parent, entrystore.get_entry(importchild))
			else:
				childiter = self.insert_entry_before(parent, sibling, entrystore.get_entry(importchild))

			baseiters.append(childiter)

			if entrystore.iter_has_child(importchild):
				self.import_entrystore(entrystore, childiter, None, importchild)

		return baseiters


	def import_entrystore_after(self, entrystore, parent, sibling):
		"""Imports entries after an existing entry"""

		return self.import_entrystore(entrystore, parent, self.iter_next(sibling))


	def import_entrystore_before(self, entrystore, parent, sibling):
		"""Imports entries before an existing entry"""

		return self.import_entrystore(entrystore, parent, sibling)


	def insert_entry_after(self, parent, sibling, data = None):
		"""Inserts an entry after a different entry"""

		iter = self.insert_after(parent, sibling)
		self.update_entry(iter, data)
		return iter


	def insert_entry_before(self, parent, sibling, data = None):
		"""Inserts an entry before a different entry"""

		iter = self.insert_before(parent, sibling)
		self.update_entry(iter, data)
		return iter


	def is_empty(self):
		"""Checks if the entrystore is empty"""

		return self.iter_n_children(None) == 0


	def remove_entry(self, iter):
		"""Recursively removes an entry"""

		while self.iter_has_child(iter):
			self.remove_entry(self.iter_children(iter))

		self.remove(iter)


	def update_entry(self, iter, data):
		"""Updates data for an entry"""

		if data == None:
			return

		self.set_value(iter, ENTRYSTORE_COL_NAME, data.get("name", ""))
		self.set_value(iter, ENTRYSTORE_COL_TYPE, data.get("type", ENTRY_FOLDER))
		self.set_value(iter, ENTRYSTORE_COL_DESC, data.get("description", ""))
		self.set_value(iter, ENTRYSTORE_COL_UPDATED, data.get("updated", 0))
		self.set_value(iter, ENTRYSTORE_COL_FIELDS, data.get("fields", {}))
		self.set_value(iter, ENTRYSTORE_COL_ICON, self.entrytypes.get_data(data.get("type", ENTRY_FOLDER))["icon"])



class DataStore(EntryStore):
	"""High-level subclass of entry store, for application use"""

	def __init__(self, entrytypes):
		"""Sets up the datastore"""

		EntryStore.__init__(self, entrytypes)
		self.changed = gtk.FALSE


	def add_entry(self, parent, data = None):
		"""Adds an entry to the data store"""

		# place inside parent if folder
		if self.get_entry_type(parent) in (ENTRY_FOLDER, None):
			iter = EntryStore.add_entry(self, parent, data)

		# place after "parent" if not folder
		else:
			iter = EntryStore.insert_entry_after(self, self.iter_parent(parent), parent, data)

		self.changed = gtk.TRUE

		return iter


	def clear(self):
		"""Removes all data from the datastore"""

		EntryStore.clear(self)
		self.changed = gtk.FALSE


	def get_entry(self, iter):
		"""Fetches data for a tree entry"""

		data = EntryStore.get_entry(self, iter)

		if data == None:
			return None

		# always return the normal, closed folder icon for folders.
		# the open folder icon is only for tree-internal use.
		if data["icon"] == revelation.ui.STOCK_FOLDER_OPEN:
			data["icon"] = revelation.ui.STOCK_FOLDER

		return data


	def remove_entry(self, iter):
		"""Removes an entry and all children"""

		parent = self.iter_parent(iter)
		EntryStore.remove_entry(self, iter)

		# collapse parent if empty
		if self.iter_n_children(parent) == 0:
			self.set_folder_state(parent, gtk.FALSE)

		self.changed = gtk.TRUE


	def set_folder_state(self, iter, open):
		"""Sets the open/closed state of a folder, changing its icon"""

		if iter == None or self.get_entry_type(iter) != ENTRY_FOLDER:
			return

		if open == gtk.TRUE:
			icon = revelation.ui.STOCK_FOLDER_OPEN
		else:
			icon = revelation.ui.STOCK_FOLDER

		self.set_value(iter, ENTRYSTORE_COL_ICON, icon)


	def update_entry(self, iter, data):
		"""Updates data for an entry"""

		if data == None:
			return

		EntryStore.update_entry(self, iter, data)

		# keep icon if current is folder-open and the type still is folder
		if data["type"] == ENTRY_FOLDER and self.get_value(iter, ENTRYSTORE_COL_ICON) == revelation.ui.STOCK_FOLDER_OPEN:
			self.set_value(iter, ENTRYSTORE_COL_ICON, revelation.ui.STOCK_FOLDER_OPEN)

		self.changed = gtk.TRUE



class EntryClipboard(EntryStore):
	"""An entry clipboard"""

	def __init__(self):

		EntryStore.__init__(self, EntryTypes())


	def copy(self, datastore, iters):
		"""Copies data from the datastore"""

		if len(iters) == 0 or iters[0] == None:
			return

		self.clear()
		datastore.export_entrystore(iters, self)
		self.emit("copy")


	def cut(self, datastore, iters):
		"""Cuts data from the datastore"""

		self.copy(datastore, iters)

		for iter in iters:
			datastore.remove_entry(iter)

		self.emit("cut")


	def paste(self, datastore, parent, sibling = None):
		"""Pastes data into the datastore"""

		if sibling == None:
			iters = datastore.import_entrystore(self, parent)
		else:
			iters = datastore.import_entrystore_after(self, parent, sibling)

		self.emit("paste")
		return iters


gobject.signal_new("copy", EntryClipboard, gobject.SIGNAL_ACTION, gobject.TYPE_BOOLEAN, ())
gobject.signal_new("cut", EntryClipboard, gobject.SIGNAL_ACTION, gobject.TYPE_BOOLEAN, ())
gobject.signal_new("paste", EntryClipboard, gobject.SIGNAL_ACTION, gobject.TYPE_BOOLEAN, ())



class UndoQueue(gobject.GObject):

	def __init__(self):

		gobject.GObject.__init__(self)

		self.queue = []
		self.actionptr = 0


	def add_action(self, name, action, data):
		"""Adds an action to the queue"""

		# remove any items later in the queue
		del self.queue[self.actionptr:]

		self.queue.append({ "name" : name, "action" : action, "data" : data })
		self.actionptr = len(self.queue)

		self.emit("can-undo", self.can_undo())
		self.emit("can-redo", self.can_undo(REDO))


	def can_undo(self, method = UNDO):
		"""Checks if the undo queue allows an undo action"""

		if method == UNDO:
			return self.actionptr > 0
		else:
			return self.actionptr < len(self.queue)


	def clear(self):
		"""Clears the undo queue"""

		self.queue = []
		self.actionptr = 0

		self.emit("can-undo", gtk.FALSE)
		self.emit("can-redo", gtk.FALSE)


	def get_data(self, method = UNDO):
		"""Returns the current undo data"""

		if method == UNDO:
			ptr = self.actionptr - 1
		else:
			ptr = self.actionptr

		if self.can_undo(method):
			return self.queue[ptr]
		else:
			return None


	def redo(self):
		"""Redoes the next action"""

		if self.can_undo(REDO):
			self.emit("redo", self.get_data(REDO))
			self.actionptr = self.actionptr + 1

		self.emit("can-undo", self.can_undo())
		self.emit("can-redo", self.can_undo(REDO))


	def undo(self):
		"""Undoes the previous action"""

		if self.can_undo():
			self.emit("undo", self.get_data())
			self.actionptr = self.actionptr - 1

		self.emit("can-undo", self.can_undo())
		self.emit("can-redo", self.can_undo(REDO))


gobject.signal_new("undo", UndoQueue, gobject.SIGNAL_ACTION, gobject.TYPE_BOOLEAN, (gobject.TYPE_PYOBJECT, ))
gobject.signal_new("redo", UndoQueue, gobject.SIGNAL_ACTION, gobject.TYPE_BOOLEAN, (gobject.TYPE_PYOBJECT, ))
gobject.signal_new("can-undo", UndoQueue, gobject.SIGNAL_ACTION, gobject.TYPE_BOOLEAN, (gobject.TYPE_BOOLEAN, ))
gobject.signal_new("can-redo", UndoQueue, gobject.SIGNAL_ACTION, gobject.TYPE_BOOLEAN, (gobject.TYPE_BOOLEAN, ))


def escape_markup(string):
	"""Escapes special characters in a string"""

	string = string.replace("&", "&amp;")
	string = string.replace("<", "&lt;")
	string = string.replace(">", "&gt;")

	return string

